import 'package:flutter/material.dart';
import 'package:knt_bloc/knt_bloc.dart';

import '../bloc/book_bloc.dart';
import '../model/book.dart';

class MainPage extends StatelessWidget {
  const MainPage({Key? key}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return BlocProvider(
      create: (context) => BookBloc(),
      child: const _Content(),
    );
  }
}

class _Content extends StatefulWidget {
  const _Content({Key? key}) : super(key: key);

  @override
  State<_Content> createState() => _ContentState();
}

class _ContentState extends State<_Content>
    with PagingAction<void, Book, BookBloc, _Content> {
  bool _usingIsolate = false;
  int _addedTimes = 0;
  int _editedTimes = 0;

  @override
  void initState() {
    super.initState();
    Future.delayed(Duration.zero, () {
      context.read<BookBloc>().add(PrepareIsolateEvent());
      Future.delayed(const Duration(milliseconds: 200), () {
        context
            .read<BookBloc>()
            .add(const BasePagingEvent<void>(isFirstRequest: true));
      });
    });
  }

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      appBar: AppBar(
        title: const Text('Top 100 Best Books'),
        actions: [
          InkWell(
            onTap: _toggleIsolate,
            child: Icon(
              _usingIsolate ? Icons.sync : Icons.sync_disabled,
              color: Colors.white,
            ),
          ),
          const SizedBox(width: 12),
          InkWell(
            onTap: _loadBooks,
            child: const Icon(
              Icons.cloud_download,
              color: Colors.white,
            ),
          ),
          const SizedBox(width: 24),
        ],
      ),
      floatingActionButton: FloatingActionButton(
        onPressed: _onAddBook,
        backgroundColor: Colors.green,
        child: const Icon(
          Icons.add,
          color: Colors.white,
          size: 24,
        ),
      ),
      body: Column(
        crossAxisAlignment: CrossAxisAlignment.stretch,
        children: [
          const _DumpViewToTestIsolatePool(),
          Expanded(
            child: PagingWidget<void, Book, BookBloc>(
              padding: const EdgeInsets.symmetric(vertical: 12),
              needQueryImmediately: false,
              onItemChanged: (action, index, item) {
                _showSnackBar(
                    '[$action] at $index = updated item ${item.title}');
              },
              emptyDataWidget: Center(
                child: ElevatedButton(
                  onPressed: () {},
                  child: const Text(
                    'Placeholder',
                    style: TextStyle(
                      color: Colors.white,
                      fontSize: 17,
                      fontWeight: FontWeight.w500,
                      height: 21 / 17,
                    ),
                  ),
                ),
              ),
              errorWidgetBuilder: (context, exception) => Center(
                child: Text(
                  'Occur an error: $exception',
                  style: const TextStyle(
                    color: Colors.blue,
                    fontSize: 17,
                    fontWeight: FontWeight.w500,
                    height: 21 / 17,
                  ),
                ),
              ),
              loadingItemBuilder: (context, index) => const Center(
                child: CircularProgressIndicator(color: Colors.blue),
              ),
              itemBuilder: (context, index, data) => _ItemBook(
                key: ValueKey(data.title),
                book: data,
                onEdit: (book) => _onEditBook(book, index),
                onDelete: _onDeleteBook,
              ),
            ),
          ),
        ],
      ),
    );
  }

  void _toggleIsolate() {
    setState(() {
      _usingIsolate = !_usingIsolate;
    });
    _showSnackBar('Isolate mode: ${_usingIsolate ? 'enabled' : 'disabled'}');
    context.read<BookBloc>().add(LoadBooksEvent(usingIsolate: _usingIsolate));
  }

  void _loadBooks() {
    _addedTimes = 0;
    _editedTimes = 0;
    context
        .read<BookBloc>()
        .add(const BasePagingEvent<void>(isFirstRequest: true));
  }

  void _onAddBook() {
    _addedTimes++;
    final newBook = Book(
      title: '[ADDED - $_addedTimes] A great read from REX',
      author: 'KienNT',
    );
    onAdd(newBook);
  }

  void _onEditBook(Book oldBook, int position) {
    _editedTimes++;
    final newBook = Book(
      title: '[EDITED - $_editedTimes] ${oldBook.title}',
      author: oldBook.author,
      country: oldBook.country,
      imageLink: oldBook.imageLink,
      language: oldBook.language,
      link: oldBook.link,
      pages: oldBook.pages,
      year: oldBook.year,
    );
    onEdit(newBook, position);
  }

  void _onDeleteBook(Book oldBook) {
    onDelete(oldBook);
  }

  void _showSnackBar(String message) {
    if (!mounted) {
      return;
    }
    ScaffoldMessenger.of(context).removeCurrentSnackBar();
    ScaffoldMessenger.of(context).showSnackBar(
      SnackBar(
        duration: const Duration(milliseconds: 2000),
        content: Text(message),
      ),
    );
  }
}

class _DumpViewToTestIsolatePool extends StatelessWidget {
  const _DumpViewToTestIsolatePool({Key? key}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return BlocBuilder<BookBloc, BaseState>(
      builder: (context, state) {
        var bookCount = 0;
        var loading = true;
        var errorMessage = '';
        if (state is ShowBooksState) {
          bookCount = state.data.length;
          loading = false;
        }
        if (state is BaseErrorState) {
          errorMessage = state.exception.additionalInfo ?? '';
        }
        return Container(
          padding: const EdgeInsets.symmetric(
            horizontal: 16.0,
            vertical: 8.0,
          ),
          color: errorMessage.isNotEmpty
              ? Colors.red
              : loading
                  ? Colors.blue
                  : Colors.green,
          child: Text(
            errorMessage.isNotEmpty
                ? 'Occur an error: [$errorMessage]'
                : 'Just a dump view to test Bloc\'s Isolate Pool'
                    '\nBook Count: ${loading ? 'loading' : bookCount}',
            style: const TextStyle(
              color: Colors.white,
              fontSize: 16,
              height: 1.5,
            ),
          ),
        );
      },
    );
  }
}

class _ItemBook extends StatelessWidget {
  const _ItemBook({
    Key? key,
    required this.book,
    this.onEdit,
    this.onDelete,
  }) : super(key: key);

  final Book book;
  final Function(Book)? onEdit;
  final Function(Book)? onDelete;

  @override
  Widget build(BuildContext context) {
    return Container(
      height: 54,
      padding: const EdgeInsets.symmetric(
        horizontal: 16,
        vertical: 4,
      ),
      child: Row(
        children: [
          const Icon(
            Icons.book_rounded,
            color: Colors.green,
            size: 24,
          ),
          const SizedBox(width: 12),
          Expanded(
            child: Column(
              crossAxisAlignment: CrossAxisAlignment.start,
              children: [
                Text(
                  book.title,
                  maxLines: 1,
                  overflow: TextOverflow.ellipsis,
                  style: const TextStyle(
                    color: Colors.blue,
                    fontSize: 17,
                    fontWeight: FontWeight.w600,
                    height: 21 / 17,
                  ),
                ),
                const SizedBox(height: 4),
                Text(
                  book.author,
                  maxLines: 1,
                  overflow: TextOverflow.ellipsis,
                  style: const TextStyle(
                    color: Colors.grey,
                    fontSize: 13,
                    fontWeight: FontWeight.w300,
                    height: 15 / 13,
                  ),
                ),
              ],
            ),
          ),
          const SizedBox(width: 12),
          InkWell(
            onTap: () => onEdit?.call(book),
            child: Container(
              width: 24,
              height: 24,
              decoration: BoxDecoration(
                color: Colors.blue,
                borderRadius: BorderRadius.circular(4),
              ),
              child: const Center(
                child: Icon(
                  Icons.edit,
                  size: 16,
                  color: Colors.white,
                ),
              ),
            ),
          ),
          const SizedBox(width: 4),
          InkWell(
            onTap: () => onDelete?.call(book),
            child: Container(
              width: 24,
              height: 24,
              decoration: BoxDecoration(
                color: Colors.red,
                borderRadius: BorderRadius.circular(4),
              ),
              child: const Center(
                child: Icon(
                  Icons.delete,
                  size: 16,
                  color: Colors.white,
                ),
              ),
            ),
          ),
        ],
      ),
    );
  }
}
