import 'dart:async';
import 'dart:io';
import 'dart:isolate';

import 'package:async/async.dart';
import 'package:bloc/bloc.dart';
import 'package:bloc_concurrency/bloc_concurrency.dart';
import 'package:connectivity_plus/connectivity_plus.dart';
import 'package:dio/dio.dart';
import 'package:meta/meta.dart';

import '../defined.dart';
import '../entity/data_paging.dart';
import '../entity/exception.dart';
import '../utils/logger.dart';

part 'events.dart';

part 'mixins.dart';

part 'states.dart';

/// Base handles normal events.
abstract class BaseBloc extends Bloc<BaseEvent, BaseState>
    with NetworkCondition {
  BaseBloc() : super(const BaseInitState()) {
    reset();
    on<BaseLoadingEvent>(
      handleBaseLoadingEvent,
      transformer: restartable(),
    );
    on<BaseErrorEvent>(
      handleBaseErrorEvent,
      transformer: restartable(),
    );
    on<CancelRunningTaskEvent>(
      handleCancelRunningTaskEvent,
      transformer: restartable(),
    );
    on<ResetBlocEvent>(
      handleResetBlocEvent,
      transformer: restartable(),
    );
  }

  CancelableOperation? _cancelableOperation;

  @override
  void onTransition(Transition<BaseEvent, BaseState> transition) {
    super.onTransition(transition);
    if (needLog) {
      logger.d('🚒🚒🚒 TRANSITION BLOC [-$runtimeType-] 🚒🚒🚒\n$transition');
    }
  }

  @override
  @mustCallSuper
  Future<void> close() async {
    await _cancelableOperation?.cancel();
    return super.close();
  }

  /// A wrapper of [on] function to ease the handling error flow.
  @protected
  void watch<E extends BaseEvent>(
    EventHandler<E, BaseState> handler, {
    EventTransformer<E>? transformer,
  }) {
    on<E>(
      (event, emit) async => _handleEventWithException(handler, event, emit),
      transformer: transformer,
    );
  }

  /// Change to loading state.
  @protected
  Future<void> handleBaseLoadingEvent(
    BaseLoadingEvent event,
    Emitter<BaseState> emit,
  ) async {
    emit(BaseLoadingState(tag: event.tag));
  }

  /// Throws error state.
  @protected
  Future<void> handleBaseErrorEvent(
    BaseErrorEvent event,
    Emitter<BaseState> emit,
  ) async {
    emit(
      BaseErrorState(
        exception: event.exception,
        tag: event.tag,
      ),
    );
  }

  /// Cancels running tasks.
  @protected
  Future<void> handleCancelRunningTaskEvent(
    CancelRunningTaskEvent event,
    Emitter<BaseState> emit,
  ) async {
    await cancelRunningTask();
  }

  /// Needs to reset data bloc, includes cancelling all running async tasks.
  @protected
  @mustCallSuper
  Future<void> handleResetBlocEvent(
    ResetBlocEvent event,
    Emitter<BaseState> emit,
  ) async {
    if (event.needToCancelRunningTask) {
      await cancelRunningTask();
    }
    reset();
    if (event.needToNotify) {
      emit(BaseResetDoneState(tag: event.tag));
    }
  }

  /// Handle exception which is thrown from mapping event-state process.
  @protected
  @mustCallSuper
  void handleError(Exception exception, {String? tag}) {
    if (exception is KntException) {
      if (needLog) {
        logger.e('❌❌ KntException: $exception ❌❌');
      }
      add(
        BaseErrorEvent(
          exception: exception,
          tag: tag,
        ),
      );
      return;
    }
    if (needLog) {
      logger.e('❌❌ Error: $exception ❌❌');
    }
    add(
      BaseErrorEvent(
        exception: convertedException(exception),
        tag: tag,
      ),
    );
  }

  @protected
  KntException convertedException(Exception exception) => KntException(
        KntIssue.undefined,
        additionalInfo: exception.toString(),
      );

  /// Reset data.
  @protected
  void reset() {}

  /// Cancel all running task in the background.
  @protected
  @mustCallSuper
  Future<void> cancelRunningTask() async {
    await _cancelableOperation?.cancel();
  }

  /// Wrap cancelable async process with connection status -
  /// [needCheckConnection] and error process if need - [needHandleError].
  @protected
  Future? cancelableFuture(
    Future inner, {
    String? tag,
    bool needCheckConnection = true,
    bool needHandleError = true,
  }) async {
    final preProcess =
        needCheckConnection ? () async => checkConnection() : null;
    final process = _processCancelableFuture(inner, preProcess: preProcess);
    return needHandleError
        ? process?.onError(
            (error, stackTrace) {
              if (error is Exception) {
                handleError(error, tag: tag);
              }
            },
          )
        : process;
  }

  /// Whether need to log bloc transition or not
  @protected
  bool get needLog => false;

  /// Create an async process which can be cancelled
  Future? _processCancelableFuture(
    Future inner, {
    Future? Function()? preProcess,
  }) async {
    await preProcess?.call();
    await _cancelableOperation?.cancel();
    _cancelableOperation = CancelableOperation.fromFuture(inner);
    return _cancelableOperation?.value;
  }

  Future<void> _handleEventWithException<E extends BaseEvent>(
    EventHandler<E, BaseState> originHandler,
    E event,
    Emitter<BaseState> emit,
  ) async {
    try {
      if (event.needCheckConnection) {
        await checkConnection();
      }
      await originHandler(event, emit);
    } on Exception catch (e) {
      handleError(e, tag: event.tag);
    }
  }
}

/// Base handles network required events & specific [DioError]
abstract class BaseApiBloc extends BaseBloc {
  @override
  KntException convertedException(Exception exception) {
    if (exception is DioError) {
      return KntException(
        convertedIssue(exception),
        code: exception.response?.statusCode,
        additionalInfo: exception.response?.statusMessage ?? exception.message,
      );
    } else {
      return super.convertedException(exception);
    }
  }

  @protected
  KntIssue convertedIssue(DioError error) {
    switch (error.type) {
      case DioErrorType.connectionTimeout:
      case DioErrorType.receiveTimeout:
        return KntIssue.timeout;
      case DioErrorType.sendTimeout:
        return KntIssue.network;
      case DioErrorType.badResponse:
        switch (error.response?.statusCode) {
          case HttpStatus.unauthorized:
            return KntIssue.unauthorized;
          case HttpStatus.badRequest:
            return KntIssue.badRequest;
          case HttpStatus.notFound:
            return KntIssue.notFound;
          case HttpStatus.methodNotAllowed:
          case HttpStatus.unprocessableEntity:
          case HttpStatus.internalServerError:
          case HttpStatus.badGateway:
          case HttpStatus.serviceUnavailable:
          case HttpStatus.gatewayTimeout:
            return KntIssue.server;
          default:
            return KntIssue.undefined;
        }
      case DioErrorType.cancel:
        return KntIssue.cancel;
      case DioErrorType.unknown:
      default:
        if (error.error is SocketException) {
          // SocketException: Failed host lookup: '***'
          // (OS Error: No address associated with hostname, errno = 7)
          return KntIssue.network;
        } else {
          return KntIssue.undefined;
        }
    }
  }
}

/// Base handle paging events
abstract class BasePagingBloc<Q, D> extends BaseApiBloc {
  BasePagingBloc() : super() {
    watch<BasePagingEvent<Q>>(
      handlePagingEvent,
      transformer: pagingTransformer,
    );
    watch<AddToPagingDataEvent<D>>(
      _handleAddToPagingDataEvent,
      transformer: droppable(),
    );
    watch<EditPagingDataEvent<D>>(
      _handleEditPagingDataEvent,
      transformer: droppable(),
    );
    watch<DeleteFromPagingDataEvent<D>>(
      _handleDeleteFromPagingDataEvent,
      transformer: droppable(),
    );
  }

  final List<D> _dataList = [];
  late int _currentOffset;
  late bool _loadedAllData;

  @override
  @mustCallSuper
  void reset() {
    _dataList.clear();
    _currentOffset = defaultOffsetPaging;
    _loadedAllData = false;
  }

  @override
  Future<void> handleBaseErrorEvent(
    BaseErrorEvent event,
    Emitter<BaseState> emit,
  ) async {
    if (_dataList.isEmpty) {
      // Error occurs at the first query
      return super.handleBaseErrorEvent(event, emit);
    }
    // Otherwise, ignore this error while loading more data
    emit(pagingStateInstance(_dataList, _currentOffset, tag: event.tag));
  }

  /// Handle query process
  @protected
  Future<BaseResponsePaging<D>> pagingInProgress(
    BaseRequestPaging<Q> pagingQuery,
  );

  /// Create instance of paging result
  @protected
  BasePagingResultState<D> pagingStateInstance(
    List<D> dataList,
    int offset, {
    String? tag,
  });

  @protected
  Future<void> handlePagingEvent(
    BasePagingEvent<Q> event,
    Emitter<BaseState> emit,
  ) async {
    // Check if all data is loaded
    final isFirstRequest = event.isFirstRequest;
    if (!isFirstRequest && _loadedAllData) {
      return;
    }
    // Check conditions
    if (pagingQueryNeedCheckConnection) {
      await checkConnection();
      // Add other conditions in the future
    }
    if (isFirstRequest) {
      reset();
    }

    // After that, start paging request
    final tag = event.tag;
    if (event.needLoading) {
      emit(BaseLoadingState(tag: tag));
    }
    final responsePaging = await cancelableFuture(
      pagingInProgress(
        BaseRequestPaging<Q>(
          queryData: event.queryData,
          limit: event.limit,
          offset: _currentOffset,
        ),
      ),
      tag: tag,
      // Already checked connection
      needCheckConnection: false,
      // Need rethrow here to spread out to outer layer
      needHandleError: false,
    );
    if (responsePaging is! BaseResponsePaging<D>) {
      throw const KntException(KntIssue.parsing);
    }
    final dataList = responsePaging.dataList;
    if (dataList.length < event.limit) {
      // Reached last page if result has less than limit
      _loadedAllData = true;
    }
    if (dataList.isEmpty && isFirstRequest) {
      // In case at first paging request we have empty data, we could stop
      emit(BaseEmptyResultState(tag: tag));
      return;
    }
    // Update data list and offset
    _dataList.addAll(dataList.where((element) => element != null));
    if (responsePaging.currentOffset == defaultInvalidOffset) {
      _currentOffset += dataList.length;
    } else {
      _currentOffset = responsePaging.currentOffset;
    }
    // Sorting
    await sort();
    // Show state
    emit(pagingStateInstance(_dataList, _currentOffset, tag: tag));
  }

  /// Place where to sort data list according to the [comparable]
  @protected
  Future<void> sort() async {
    if (comparable == null) {
      return;
    }
    final unsortedData = _dataList;
    final comparableFn = comparable;
    final results = await Isolate.run<List<D>>(() {
      unsortedData.sort(comparableFn);
      return unsortedData;
    });
    _dataList
      ..clear()
      ..addAll(results);
  }

  @protected
  D? find(bool Function(D) filter) {
    for (final data in _dataList) {
      if (filter(data)) {
        return data;
      }
    }
    return null;
  }

  @protected
  int indexOf(D data) => _dataList.indexOf(data);

  /// Override this getter when we want to sort out result.
  @protected
  CompareFunction<D>? get comparable => null;

  /// Override this getter when we want to check internet connection before
  /// querying.
  @protected
  bool get pagingQueryNeedCheckConnection => false;

  @protected
  EventTransformer<BasePagingEvent<Q>> get pagingTransformer => sequential();

  Future<void> _handleAddToPagingDataEvent(
    AddToPagingDataEvent<D> event,
    Emitter<BaseState> emit,
  ) async {
    final newData = event.newData;
    _currentOffset++;
    _dataList.add(newData);
    await sort();
    final updatedPosition = _dataList.indexOf(newData);
    emit(
      ChangedPagingDataState(
        action: PagingDataAction.create,
        data: newData,
        newPosition: event.focusOnUpdatedData ? updatedPosition : 0,
        focusToNewPosition: event.focusOnUpdatedData,
        tag: event.tag,
      ),
    );
    emit(pagingStateInstance(_dataList, _currentOffset, tag: event.tag));
  }

  Future<void> _handleEditPagingDataEvent(
    EditPagingDataEvent<D> event,
    Emitter<BaseState> emit,
  ) async {
    final currentPosition = event.currentPosition;
    if (currentPosition < 0 || currentPosition >= _dataList.length) {
      return;
    }
    final updatedData = event.updatedData;
    _dataList[currentPosition] = updatedData;
    await sort();
    final updatedPosition = _dataList.indexOf(updatedData);
    emit(
      ChangedPagingDataState(
        action: PagingDataAction.edit,
        data: updatedData,
        newPosition: event.focusOnUpdatedData ? updatedPosition : 0,
        focusToNewPosition: event.focusOnUpdatedData,
        tag: event.tag,
      ),
    );
    emit(pagingStateInstance(_dataList, _currentOffset, tag: event.tag));
  }

  Future<void> _handleDeleteFromPagingDataEvent(
    DeleteFromPagingDataEvent<D> event,
    Emitter<BaseState> emit,
  ) async {
    final oldData = event.data;
    final oldPosition = _dataList.indexOf(oldData);
    if (oldPosition > -1) {
      _currentOffset--;
    }
    _dataList.remove(oldData);
    emit(
      ChangedPagingDataState(
        action: PagingDataAction.delete,
        data: oldData,
        newPosition: _dataList.isEmpty
            ? -1
            : (event.focusOnUpdatedData ? oldPosition : 0),
        focusToNewPosition: event.focusOnUpdatedData,
        tag: event.tag,
      ),
    );
    emit(pagingStateInstance(_dataList, _currentOffset, tag: event.tag));
  }
}

typedef CompareFunction<T> = int Function(T, T);
